from colorama import Fore, Style

GREEN_PLUS = f"{Fore.GREEN}[+]{Style.RESET_ALL}"
GREEN_EXCLAIM = f"{Fore.GREEN}[!]{Style.RESET_ALL}"
YELLOW_DASH = f"{Fore.YELLOW}[-]{Style.RESET_ALL}"
YELLOW_EXCLAIM = f"{Fore.YELLOW}[!]{Style.RESET_ALL}"
RED_DASH = f"{Fore.RED}[-]{Style.RESET_ALL}"
RED_EXCLAIM = f"{Fore.RED}[!]{Style.RESET_ALL}"
BRIGHT_DASH = f"{Style.BRIGHT}-{Style.RESET_ALL}"
